/*!---------------------------------------------------------
* Copyright (C) Microsoft Corporation. All rights reserved.
*--------------------------------------------------------*/
/// <reference path="./declare/node.d.ts" />
/// <reference path="./declare/atom-renderer.d.ts" />
'use strict';
var remote = require('remote');
var ipc = require('ipc');
var path = require('path');
var IndexBase = (function () {
    function IndexBase(name, enableDeveloperTools) {
        this.name = name;
        this.enableDeveloperTools = enableDeveloperTools;
        this.registerListeners();
    }
    IndexBase.prototype._uriFromPath = function (_path) {
        var pathName = path.resolve(_path).replace(/\\/g, '/');
        if (pathName.length > 0 && pathName.charAt(0) !== '/') {
            pathName = '/' + pathName;
        }
        return encodeURI('file://' + pathName);
    };
    IndexBase.prototype.registerListeners = function () {
        var _this = this;
        // Devtools & reload support
        if (this.enableDeveloperTools) {
            var extractKey = function (e) {
                return [
                    e.ctrlKey ? 'ctrl-' : '',
                    e.metaKey ? 'meta-' : '',
                    e.altKey ? 'alt-' : '',
                    e.shiftKey ? 'shift-' : '',
                    e.keyCode
                ].join('');
            };
            // mac: Cmd-Alt-I, rest: Ctrl-Shift-I
            var TOGGLE_DEV_TOOLS_KB = (process.platform === 'darwin' ? 'meta-alt-73' : 'ctrl-shift-73');
            // mac: Cmd-R, rest: Ctrl-R
            var RELOAD_KB = (process.platform === 'darwin' ? 'meta-82' : 'ctrl-82');
            window.addEventListener('keydown', function (e) {
                var key = extractKey(e);
                if (key === TOGGLE_DEV_TOOLS_KB) {
                    remote.getCurrentWindow().toggleDevTools();
                }
                else if (key === RELOAD_KB) {
                    ipc.send('ticino:reloadWindow', remote.getCurrentWindow().id);
                }
            });
        }
        // Workaround to prevent page navigation when backspace is pressed
        window.addEventListener('keydown', function (e) {
            if (e.which !== 8) {
                return; // only interested in backspace
            }
            var targetTagName = e.target && e.target.tagName && e.target.tagName.toLowerCase();
            var targetReadOnly = e.target && e.target.readOnly;
            if (targetTagName && (targetTagName === 'input' || targetTagName === 'textarea') && !targetReadOnly) {
                return; // allow backspace in editable input/textarea
            }
            e.preventDefault();
        });
        process.on('uncaughtException', function (err) { return _this.onUncaughtException(err); });
    };
    IndexBase.prototype.onUncaughtException = function (err) {
        console.error('[uncaught exception in ' + this.name + ']: ', err);
        if (err.stack) {
            console.error(err.stack);
        }
        if (this.enableDeveloperTools) {
            remote.getCurrentWindow().openDevTools();
            remote.getCurrentWindow().show();
        }
    };
    IndexBase.prototype.createScript = function (src, onload) {
        var script = document.createElement('script');
        script.src = src;
        script.addEventListener('load', onload);
        var head = document.getElementsByTagName('head')[0];
        head.insertBefore(script, head.lastChild);
    };
    IndexBase.prototype.parseURLQueryArgs = function () {
        var result = {};
        var search = window.location.search;
        if (search) {
            var params = search.split(/[?&]/);
            for (var i = 0; i < params.length; i++) {
                var param = params[i];
                if (param) {
                    var keyValue = param.split('=');
                    if (keyValue.length === 2) {
                        result[keyValue[0]] = decodeURIComponent(keyValue[1]);
                    }
                }
            }
        }
        return result;
    };
    return IndexBase;
})();
exports.IndexBase = IndexBase;
