_ = require 'underscore'

fs = require 'fs'
Grammar = require './grammar'
NullGrammar = require './null-grammar'

# Extended: Registry containing one or more grammars.
module.exports =
class GrammarRegistry
  constructor: (options={}) ->
    @maxTokensPerLine = options.maxTokensPerLine ? Infinity
    @grammars = []
    @grammarsByScopeName = {}
    @injectionGrammars = []
    @grammarOverridesByPath = {}
    @scopeIdCounter = -1
    @idsByScope = {}
    @scopesById = {}

    @nullGrammar = new NullGrammar(this)
    @addGrammar(@nullGrammar)

  ###
  Section: Managing Grammars
  ###

  # Public: Get all the grammars in this registry.
  #
  # Returns a non-empty {Array} of {Grammar} instances.
  getGrammars: ->
    _.clone(@grammars)

  # Public: Get a grammar with the given scope name.
  #
  # * `scopeName` A {String} such as `"source.js"`.
  #
  # Returns a {Grammar} or undefined.
  grammarForScopeName: (scopeName) ->
    @grammarsByScopeName[scopeName]

  # Public: Add a grammar to this registry.
  #
  # * `grammar` The {Grammar} to add. This should be a value previously returned
  #   from {::readGrammarSync}.
  #
  # Returns the {Grammar}.
  addGrammar: (grammar) ->
    @grammars.push(grammar)
    @grammarsByScopeName[grammar.scopeName] = grammar
    @injectionGrammars.push(grammar) if grammar.injectionSelector?
    undefined

  removeGrammar: (grammar) ->
    index = @grammars.indexOf(grammar)
    @grammars.splice(index, 1) if index >= 0
    
    delete @grammarsByScopeName[grammar.scopeName]
    
    index = @injectionGrammars.indexOf(grammar)
    @injectionGrammars.splice(index, 1) if index >= 0
    undefined

  # Public: Remove the grammar with the given scope name.
  #
  # * `scopeName` A {String} such as `"source.js"`.
  #
  # Returns the removed {Grammar} or undefined.
  removeGrammarForScopeName: (scopeName) ->
    grammar = @grammarForScopeName(scopeName)
    @removeGrammar(grammar) if grammar?
    grammar

  # Public: Read a grammar synchronously but don't add it to the registry.
  #
  # * `grammarPath` A {String} absolute file path to a grammar file.
  #
  # Returns a {Grammar}.
  readGrammarSync: (grammarPath) ->
    contents = fs.readFileSync(grammarPath.replace(/\.cson$/, '.json'), 'utf8')
    grammar = null
    if contents.trim().length isnt 0
      grammar = JSON.parse(contents)
      
    if typeof grammar.scopeName is 'string' and grammar.scopeName.length > 0
      @createGrammar(grammarPath, grammar)
    else
      throw new Error("Grammar missing required scopeName property: #{grammarPath}")

  # Public: Read a grammar synchronously and add it to this registry.
  #
  # * `grammarPath` A {String} absolute file path to a grammar file.
  #
  # Returns a {Grammar}.
  loadGrammarSync: (grammarPath) ->
    grammar = @readGrammarSync(grammarPath)
    @addGrammar(grammar)
    grammar

  # Public: Get the grammar override for the given file path.
  #
  # * `filePath` A {String} file path.
  #
  # Returns a {Grammar} or undefined.
  grammarOverrideForPath: (filePath) ->
    @grammarOverridesByPath[filePath]

  # Public: Set the grammar override for the given file path.
  #
  # * `filePath` A non-empty {String} file path.
  # * `scopeName` A {String} such as `"source.js"`.
  #
  # Returns a {Grammar} or undefined.
  setGrammarOverrideForPath: (filePath, scopeName) ->
    if filePath
      @grammarOverridesByPath[filePath] = scopeName

  # Public: Remove the grammar override for the given file path.
  #
  # * `filePath` A {String} file path.
  #
  # Returns undefined.
  clearGrammarOverrideForPath: (filePath) ->
    delete @grammarOverridesByPath[filePath]
    undefined

  # Public: Remove all grammar overrides.
  #
  # Returns undefined.
  clearGrammarOverrides: ->
    @grammarOverridesByPath = {}
    undefined

  # Public: Select a grammar for the given file path and file contents.
  #
  # This picks the best match by checking the file path and contents against
  # each grammar.
  #
  # * `filePath` A {String} file path.
  # * `fileContents` A {String} of text for the file path.
  #
  # Returns a {Grammar}, never null.
  selectGrammar: (filePath, fileContents) ->
    _.max @grammars, (grammar) -> grammar.getScore(filePath, fileContents)

  startIdForScope: (scope) ->
    unless id = @idsByScope[scope]
      id = @scopeIdCounter
      @scopeIdCounter -= 2
      @idsByScope[scope] = id
      @scopesById[id] = scope
    id

  endIdForScope: (scope) ->
    @startIdForScope(scope) - 1

  scopeForId: (id) ->
    if (id % 2) is -1
      @scopesById[id] # start id
    else
      @scopesById[id + 1] # end id

  createGrammar: (grammarPath, object) ->
    object.maxTokensPerLine ?= @maxTokensPerLine
    grammar = new Grammar(this, object)
    grammar.path = grammarPath
    grammar

  decodeTokens: (lineText, tags, scopeTags = [], fn) ->
    offset = 0
    scopeNames = scopeTags.map (tag) => @scopeForId(tag)

    tokens = []
    for tag, index in tags
      # positive numbers indicate string content with length equaling the number
      if tag >= 0
        token = {
          value: lineText.substring(offset, offset + tag)
          scopes: scopeNames.slice()
        }
        token = fn(token, index) if fn?
        tokens.push(token)
        offset += tag

      # odd negative numbers are begin scope tags
      else if (tag % 2) is -1
        scopeTags.push(tag)
        scopeNames.push(@scopeForId(tag))

      # even negative numbers are end scope tags
      else
        scopeTags.pop()
        expectedScopeName = @scopeForId(tag + 1)
        poppedScopeName = scopeNames.pop()
        unless poppedScopeName is expectedScopeName
          throw new Error("Expected popped scope to be #{expectedScopeName}, but it was #{poppedScopeName}")

    tokens

