/*!---------------------------------------------------------
* Copyright (C) Microsoft Corporation. All rights reserved.
*--------------------------------------------------------*/
/// <reference path="../declare/atom-browser.d.ts" />
/// <reference path="../declare/node.d.ts" />
/// <reference path="../declare/winreg.d.ts" />
'use strict';
var fs = require('fs');
var path = require('path');
var cp = require('child_process');
var app = require('app');
function zip(arr1, arr2) {
    return arr1.map(function (v, i) { return [v, arr2[i]]; });
}
function spawn(cmd, args, cb) {
    try {
        var child = cp.spawn(cmd, args);
    }
    catch (err) {
        process.nextTick(function () { return cb(err); });
        return;
    }
    var error = null;
    var buffers = [];
    child.stdout.on('data', function (b) { return buffers.push(b); });
    child.on('error', function (err) { return error = (error || err); });
    child.on('close', function (code) {
        if (code !== 0) {
            error = error || new Error('Command returned ' + code);
            error.code = code;
        }
        cb(error, Buffer.concat(buffers).toString('utf8'));
    });
}
var Registry = (function () {
    function Registry() {
    }
    Registry.prototype.query = function (key, valueName, cb) {
        var args = ['QUERY', key];
        args = args.concat(valueName ? ['/v', valueName] : ['/ve']);
        this.spawn(args, function (err, stdout) {
            if (err) {
                if (err.code === 1) {
                    if (/ERROR: Registry editing has been disabled by your administrator\./.test(stdout)) {
                        return cb(new Error('Failed to query due to lack of permissions'));
                    }
                    return cb(null, '');
                }
                return cb(err);
            }
            var lines = stdout.split(/[\r\n]+/).filter(function (l) { return !!l; });
            var lastLine = lines[lines.length - 1];
            if (!lastLine) {
                return cb(new Error('failed to query'));
            }
            var regexp = new RegExp("^    " + valueName + "    [\\w_]+    (.*)$");
            var match = regexp.exec(lastLine);
            if (!match) {
                return cb(new Error('failed to query'));
            }
            cb(null, match[1]);
        });
    };
    Registry.prototype.add = function (key, valueName, data, cb) {
        var args = ['ADD', key];
        args = args.concat(valueName ? ['/v', valueName] : ['/ve']);
        args.push('/t', 'REG_SZ', '/d', data, '/f');
        this.spawn(args, cb);
    };
    Registry.prototype.delete = function (key, cb) {
        this.spawn(['DELETE', key, '/f'], cb);
    };
    Registry.prototype.spawn = function (args, cb) {
        spawn(Registry.ExecutablePath, args, cb);
    };
    Registry.ExecutablePath = process.env.SystemRoot ? path.join(process.env.SystemRoot, 'System32', 'reg.exe') : 'reg.exe';
    return Registry;
})();
var SetX = (function () {
    function SetX() {
    }
    SetX.prototype.set = function (key, value, cb) {
        this.spawn([key, value], cb);
    };
    SetX.prototype.spawn = function (args, cb) {
        spawn(SetX.ExecutablePath, args, cb);
    };
    SetX.ExecutablePath = process.env.SystemRoot ? path.join(process.env.SystemRoot, 'System32', 'setx.exe') : 'setx.exe';
    return SetX;
})();
function asyncMap(array, fn, cb) {
    var loop = function (i) {
        if (i >= array.length) {
            return cb(null);
        }
        fn(array[i], function (err) {
            if (err) {
                return cb(err);
            }
            loop(i + 1);
        });
    };
    loop(0);
}
var Squirrel = (function () {
    function Squirrel(execPath, url) {
        this.url = url;
        this.appPath = path.resolve(execPath, '..');
        this.rootPath = path.resolve(this.appPath, '..');
        this.updateExePath = path.resolve(this.rootPath, 'Update.exe');
        this.exeName = path.basename(execPath);
    }
    Object.defineProperty(Squirrel.prototype, "isAvailable", {
        get: function () {
            return fs.existsSync(this.updateExePath);
        },
        enumerable: true,
        configurable: true
    });
    Squirrel.prototype.download = function (cb) {
        this.spawn(['--download', this.url], function (err, out) {
            if (err) {
                return cb(err);
            }
            try {
                var result = JSON.parse(out.trim().split('\n').pop());
                cb(null, result.releasesToApply.pop());
            }
            catch (err) {
                cb(err);
            }
        });
    };
    Squirrel.prototype.update = function (cb) {
        this.spawn(['--update', this.url], cb);
    };
    Squirrel.prototype.createShortcut = function (cb) {
        this.spawn(['--createShortcut', this.exeName], cb);
    };
    Squirrel.prototype.removeShortcut = function (cb) {
        this.spawn(['--removeShortcut', this.exeName], cb);
    };
    Squirrel.prototype.processStart = function (cb) {
        this.spawn(['--processStart', this.exeName], cb);
    };
    Squirrel.prototype.spawn = function (args, cb) {
        cb = cb || (function () { return null; });
        if (!this.isAvailable) {
            return cb(null, '');
        }
        var child = cp.spawn(this.updateExePath, args, { detached: true });
        var error = null;
        var stdout = '';
        child.stdout.setEncoding('utf8');
        child.stdout.on('data', function (d) { return stdout += d; });
        child.on('error', function (err) { return error = (error || err); });
        child.on('close', function (code) {
            if (code !== 0) {
                error = error || new Error('Command returned ' + code);
            }
            cb(error, stdout);
        });
    };
    return Squirrel;
})();
exports.Squirrel = Squirrel;
var squirrel = new Squirrel(process.execPath, null);
var registryKeys = [
    'HKCU\\Software\\Classes\\*\\shell\\Ticino',
    'HKCU\\Software\\Classes\\directory\\shell\\Ticino',
    'HKCU\\Software\\Classes\\directory\\background\\shell\\Ticino'
];
function installMenuAction(key, arg, cb) {
    var reg = new Registry();
    reg.add(key, '', 'Open with Code', function (err) {
        if (err) {
            return cb(err);
        }
        reg.add(key, 'Icon', process.execPath, function (err) {
            if (err) {
                return cb(err);
            }
            reg.add(key + '\\command', '', "\"" + process.execPath + "\" \"" + arg + "\"", cb);
        });
    });
}
// Nuget packaging excludes .nuspec files. So, at build time
// the .nuspec files are renamed to something else. At install
// time, those files should be renamed back to .nuspec.
function applyRenameJson() {
    var files = null;
    try {
        files = require('./rename.json');
    }
    catch (e) {
        return;
    }
    var root = path.dirname(__dirname);
    files.forEach(function (i) { return fs.renameSync(path.join(root, i.rename), path.join(root, i.original)); });
    fs.unlinkSync(path.join(__dirname, 'rename.json'));
}
var cmdContents = [
    '@echo off',
    'set VSCODE_CWD=%cd%',
    '"%~dp0\\..\\Update.exe" --processStart Code.exe -a="%*"'
].join('\r\n');
function addBinDirectoryToPATH(cb) {
    var root = path.resolve(path.join(__dirname, '..', '..', '..', '..'));
    var bin = path.join(root, 'bin');
    var cmd = path.join(bin, 'code.cmd');
    try {
        // create bin directory
        fs.mkdirSync(bin);
    }
    catch (e) {
    }
    try {
        // create bin shortcut
        fs.writeFileSync(cmd, cmdContents);
    }
    catch (e) {
        return cb(e);
    }
    var registry = new Registry();
    registry.query('HKCU\\Environment', 'PATH', function (err, value) {
        if (err) {
            return cb(err);
        }
        var paths = value.split(';');
        // do nothing if bin is already in PATH
        if (paths.indexOf(bin) > -1) {
            return cb();
        }
        // add bin directory to PATH
        paths.push(bin);
        var setx = new SetX();
        setx.set('PATH', paths.join(';'), cb);
    });
}
function install(cb) {
    applyRenameJson();
    addBinDirectoryToPATH(function (err) {
        if (err) {
            return cb(err);
        }
        squirrel.createShortcut(function (err) {
            if (err) {
                return cb(err);
            }
            asyncMap(zip(registryKeys, ['%1', '%1', '%V']), function (el, cb) { return installMenuAction(el[0], el[1], cb); }, cb);
        });
    });
}
function uninstallMenuAction(key, cb) {
    var reg = new Registry();
    reg.delete(key, cb);
}
function removeBinDirectoryFromPATH(cb) {
    var root = path.resolve(path.join(__dirname, '..', '..', '..', '..'));
    var bin = path.join(root, 'bin');
    var registry = new Registry();
    registry.query('HKCU\\Environment', 'PATH', function (err, value) {
        var paths = value.split(';');
        var index = paths.indexOf(bin);
        // do nothing if bin is not in PATH
        if (index === -1) {
            return cb();
        }
        // remove bin directory from PATH
        paths.splice(index, 1);
        var setx = new SetX();
        setx.set('PATH', paths.join(';'), cb);
    });
}
function uninstall(cb) {
    removeBinDirectoryFromPATH(function (err) {
        if (err) {
            return cb(err);
        }
        squirrel.removeShortcut(function (err) {
            if (err) {
                return cb(err);
            }
            asyncMap(registryKeys, uninstallMenuAction, cb);
        });
    });
}
function handleArgv() {
    var quit = function () { return app.quit(); };
    var commands = process.argv.map(function (a) { return /^--squirrel-(\w+)$/.exec(a); }).filter(function (m) { return !!m; }).map(function (m) { return m[1]; });
    switch (commands[0]) {
        case 'install':
            install(quit);
            return true;
        case 'updated':
            install(quit);
            return true;
        case 'uninstall':
            uninstall(quit);
            return true;
        case 'obsolete':
            quit();
            return true;
    }
    return false;
}
exports.handleArgv = handleArgv;
